module Main where
import Control.Concurrent
import Control.Monad
import System.IO

import Control.Parallel.Strategies

-- data MVar a

-- newEmptyMVar :: IO (MVar a)
-- newMVar      :: a -> IO (MVar a)
-- takeMVar     :: MVar a -> IO a
-- putMVar      :: MVar a -> a -> IO ()

quadrado :: Double -> Double
quadrado = (**2)

trigo :: Double -> Double
trigo x = (sin x + x / pi) * tan x

trap :: (Double -> Double) -> Double -> Double -> Int -> Double -> Double
trap f a b n h =
  h * (somaExtremidades + somaInterna)
  where
    somaExtremidades = (f a + f b) / 2
    somaInterna = sum [f $ a + fromIntegral k * h | k <- [1..n-1]]

area :: (Double -> Double) -> Double -> Double -> Int -> Double
area f a b n =
  trap f a b n h
  where
    h = (b - a) / fromIntegral n

newtype ASync a = Async (MVar a)

async :: IO a -> IO (ASync a)
async acao = do
  var <- newEmptyMVar
  forkIO (do r <- acao; putMVar var r)
  return $ Async var

wait :: ASync a -> IO a
wait (Async var) = readMVar var

-- Eval a

-- runEval :: Eval a -> a
-- rpar :: a -> Eval a
-- rseq :: a -> Eval a

areaPar ::  (Double -> Double) -> Double -> Double -> Int -> Int -> Double
areaPar f a b n t =
  sum $ runEval areas
  where
    evs = map (rpar . localArea) [0..t-1]
    areas = mapM (>>= rseq) evs
    h = (b - a) / fromIntegral n
    localN = n `div` t
    localA rk = a + rk * fromIntegral localN * h
    localB rk = localA rk + fromIntegral localN * h
    localArea rk = let rk' = fromIntegral rk in
      area f (localA rk') (localB rk') localN

-- v0 <- rpar e0
-- v1 <- rpar e1
-- recheio
-- que faz um monte coisa
-- v0'<- rseq v0
-- v1'<- rseq v1

areaCon :: (Double -> Double) -> Double -> Double -> Int -> Int -> IO Double
areaCon f a b n t = do
  asyncs <- mapM (async . localArea) [0..t-1]
  areas <- mapM wait asyncs
  return $ sum areas
  where
    h = (b - a) / fromIntegral n
    localN = n `div` t
    localA rk = a + rk * fromIntegral localN * h
    localB rk = localA rk + fromIntegral localN * h
    localArea rk = let rk' = fromIntegral rk in
      return $ area f (localA rk') (localB rk') localN

main :: IO ()
main = do
  -- hSetBuffering stdout NoBuffering
  -- m <- newEmptyMVar
  -- forkIO $ do threadDelay 500 ; putMVar m "um"
  -- forkIO $ putMVar m "dois"
  -- v1 <- takeMVar m
  -- print v1
  -- v2 <- takeMVar m
  -- print v2

  -- forkIO $ replicateM_ 1000 (putChar 'A')
  -- replicateM_ 1000 (putChar 'B')

  let a = 0
      b = 3456
      n = 10000000
      f = trigo
      as = area f a b n
  putStr "Área sequencial: "
  print as

  ca <- areaCon f a b n 2
  putStr "Área concorrente: "
  print ca

  let pa = areaPar f a b n 2
  putStr "Área paralela: "
  print pa
